"""Tests for handling mathematica's binary object representation.

This is primarily useful for dealing with compressed fields.
"""


from pyecl.models import ConstellationFieldParser


def test_decompress_integer():
    """Compress[4] -> 1:eJxTTMoPymRhYGAAAAtUAbI="""
    assert ConstellationFieldParser().decompress_field(b"1:eJxTTMoPymRhYGAAAAtUAbI=") == "4"


def test_decompress_large_integer():
    """
    In[1]:= Compress[90348590834890590349058038945]
    Out[1]:= 1:eJxTTMoP8pRlYGCwNDA2sTC1NLAAUpYGpiAukLIwMLawNDEFAJVXB6k=
    """
    assert (
        ConstellationFieldParser().decompress_field(
            b"1:eJxTTMoP8pRlYGCwNDA2sTC1NLAAUpYGpiAukLIwMLawNDEFAJVXB6k="
        )
        == "90348590834890590349058038945"
    )


def test_decompress_simpleFloat():
    """Compress[4.0] -> "1:eJxTTMoPKmIAAwEHABKtAgc=" """
    assert ConstellationFieldParser().decompress_field(b"1:eJxTTMoPKmIAAwEHABKtAgc=") == "4.0"


def test_decompress_complexFloat():
    """Compress[3.432643] -> "1:eJxTTMoPKlLeX9fBW8btAAAgJgRt" """
    assert (
        ConstellationFieldParser().decompress_field(b"1:eJxTTMoPKlLeX9fBW8btAAAgJgRt")
        == "3.432643"
    )


def test_decompress_string():
    """Compress["Hi There!"] -> "1:eJxTTMoPCuZkYGDwyFQIyUgtSlUEACgoBIs=" """
    assert (
        ConstellationFieldParser().decompress_field(b"1:eJxTTMoPCuZkYGDwyFQIyUgtSlUEACgoBIs=")
        == "Hi There!"
    )


def test_decompress_symbol():
    """Compress[Login] -> "1:eJxTTMoPKmZlYGDwyU/PzAMAGegDtg==" """
    assert (
        ConstellationFieldParser().decompress_field(b"1:eJxTTMoPKmZlYGDwyU/PzAMAGegDtg==")
        == "Login"
    )


def test_decompress_empty_list():
    """Compress[{}] -> "1:eJxTTMoPSmMAgmIWIOGTWVwCACIoA74=" """
    assert (
        ConstellationFieldParser().decompress_field(b"1:eJxTTMoPSmMAgmIWIOGTWVwCACIoA74=")
        == "List[]"
    )


def test_decompress_list_of_numbers():
    """Compress[{1, 2}] -> "1:eJxTTMoPSmNiYGAoZgESPpnFJZmMQEYmSAgATfoElQ==" """
    assert (
        ConstellationFieldParser().decompress_field(
            b"1:eJxTTMoPSmNiYGAoZgESPpnFJZmMQEYmSAgATfoElQ=="
        )
        == "List[1, 2]"
    )


def test_decompress_large_matrix():
    """Large matricies of real numbers get compresed into a special format, which we test here"""
    with open("tests/decompress-large-matrix.dat", "rb") as handle:
        parsed_field = ConstellationFieldParser().parse_variable_unit_field(handle.read())

    # The result should look like:
    # {{0.min,-87.529Lsus},{0.016667min,-96.7016Lsus},<<357>>,{5.98333min,1742.05Lsus},{6.min,1009.1Lsus}}

    assert len(parsed_field) == 361
    assert parsed_field[0][0].value == 0.0
    assert parsed_field[0][0].unit == "Minutes"
    assert parsed_field[0][1].value == -87.528984
    assert parsed_field[0][1].unit == "IndependentUnit[Lsus]"
    assert parsed_field[-1][0].value == 6.0
    assert parsed_field[-1][0].unit == "Minutes"
    assert parsed_field[-1][1].value == 1009.098328
    assert parsed_field[-1][1].unit == "IndependentUnit[Lsus]"



import base64
import json
import os
import pathlib
import tempfile
from http import HTTPStatus

import backoff
import boto3
import requests
import rollbar
import xmltodict

from .exceptions import (
    ConstellationException,
    ConstellationInvalidMethodException,
    ConstellationInvalidResponseException,
    ConstellationMissingFieldException,
    ConstellationNotLoggedInException,
    ConstellationObjectDoesNotExistException,
)
from .models import CloudFileArgs, ConstellationFieldParser
from .utils import md5_hash_file, parts_to_xml


class Constellation:
    _UPLOAD_PART_SIZE = 10_000_000  # 10 MB
    _MULTIPART_CLOUDFILE_THRESHOLD = 100_000_000  # 100 MB

    def __init__(self, auth_token="", host=""):
        self._host = host or "https://constellation.emeraldcloudlab.com"
        self._auth_token = auth_token

        self._temp_directory = None

    # Retry decorator for requests failures (like timeout or connection error)
    _retry_requests_decorator = backoff.on_exception(
        wait_gen=backoff.expo,
        exception=(requests.exceptions.Timeout, requests.exceptions.ConnectionError),
        max_tries=5,
    )

    # Configuration related functions
    def host(self):
        """Return the url of the instance constellation that the client is connecting to"""
        return self._host

    def download_directory(self):
        """Returns where cloud file downloads are stored"""
        if self._temp_directory is None:
            self._temp_directory = tempfile.TemporaryDirectory()

        return self._temp_directory.name

    # Identity requests
    @_retry_requests_decorator
    def login(self, username, password):
        """Return the auth token token for the supplied username and password"""
        auth_token = self._send_and_validate_request(
            "/ise/signintoken",
            {"username": username, "password": password},
            "AuthToken",
            requires_auth_token=False,
        )
        self.set_auth_token(auth_token)
        return self.me()

    def is_logged_in(self):
        return self._auth_token != ""

    def set_auth_token(self, auth_token):
        """Set the auth token used by the constellation client"""
        self._auth_token = auth_token

    @_retry_requests_decorator
    def me(self):
        """Return the response from the /ise/me endpoint"""
        return self._send_and_validate_request("/ise/me", None, None, method="GET")

    # Searching objects
    def search(self, ecl_type, query):
        """Perform a simple search on a given type"""
        payload = {
            "queries": [
                {
                    "clauses": [
                        {
                            "Types": [ecl_type],
                            "Query": query,
                        }
                    ],
                    "SubTypes": True,
                    "date": None,
                    "ignore_time": False,
                }
            ],
        }

        return self._send_and_validate_request(
            "/obj/search",
            payload,
            field_to_retrieve="Results",
        )

    # Downloading objects
    def download(self, objects, field_string="All", auto_download_cloud_files=False):
        """Download the supplied fields off the supplied object or objects.

        Note that fields can traverse across links using []'s.  For example:

        download(self.me()["ID"], "FinancingTeams[Name]")

        will download the status field off the protocol link of the supplied
        object.
        """
        result = None
        collapse_fields = False
        collapse_objects = False

        # Normalize the object and fields so they're always lists,
        # even if they're passed as single things.  This allows to
        # handle things like download("obj1", "Name") and
        # download([obj1], ["Name"]) in the same interface
        if not isinstance(field_string, list):
            field_string = [field_string]
            collapse_fields = True

        if not isinstance(objects, list):
            objects = [objects]
            collapse_objects = True

        result = self._download_objects(objects, field_string, auto_download_cloud_files)

        # Unwind the normalization of fields and object ids into lists
        if collapse_fields and collapse_objects:
            return result[0][0]
        if collapse_fields:
            return [x[0] for x in result]
        if collapse_objects:
            return result[0]

        return result

    def _download_objects(self, objects, field_string, auto_download_cloud_files):
        """Download multiple objects in a single request"""

        # Build the request
        traversals = self._parse_download_field_string(field_string)
        obj_requests = []
        for obj in objects:
            obj_request = {"object": {"id": obj.id}, "limit": 0, "subfield_info": True}
            if traversals:
                obj_request["traversals"] = traversals
            obj_requests.append(obj_request)

        # Send the request to constellation
        responses = self._send_and_validate_request(
            "/obj/download",
            {"requests": obj_requests},
            "responses",
        )
        if len(responses) != len(objects):
            raise ConstellationInvalidResponseException(
                f"Expected exactly {len(objects)} response from /obj/download", responses
            )

        # Parse the json results into objects
        response_objects = {}
        field_parser = ConstellationFieldParser(self.download_blob_ref)
        for response in responses:
            (
                sub_response_objects,
                sub_response_field_summaries,
            ) = self._parse_field_values_from_response(response)
            response_objects.update(sub_response_objects)
            for ecl_type, field_summaries in sub_response_field_summaries.items():
                field_parser.add_summaries_for_type(field_summaries, ecl_type)

        # Map the objects into a formatted result based on the fields string
        parsed_results = self._parse_results_from_response(response_objects, objects, traversals, field_parser)

        # If requested, go ahead and download all the cloud files encountered
        if auto_download_cloud_files:
            for cloud_file in field_parser.cloud_files_encountered():
                self.download_cloud_file(cloud_file)

        # return the results
        return parsed_results

    def _parse_field_values_from_response(self, response):
        """Parse out the field values contained in the response"""
        response_objects = {}
        field_summaries = {}
        response_fields = response.get("fields", {})
        response_field_summaries = response.get("summaries", {})
        ecl_type = response.get("resolved_object", {}).get("type")
        object_id = response.get("resolved_object", {}).get("id")

        if ecl_type is None or object_id is None:
            raise ConstellationObjectDoesNotExistException(response.get("object", {}).get("id"))

        response_objects[object_id] = {"type": ecl_type, "id": object_id}
        response_objects[object_id].update(response_fields)
        if ecl_type not in field_summaries:
            field_summaries[ecl_type] = {}
        field_summaries[ecl_type].update(response_field_summaries)
        for traversed_object in response.get("traversed_objects", []):
            (
                traversed_response_objects,
                traversed_field_summaries,
            ) = self._parse_field_values_from_response(traversed_object)
            response_objects.update(traversed_response_objects)
            field_summaries.update(traversed_field_summaries)

        return response_objects, field_summaries

    def _parse_download_field_string(self, field_string):
        """Parse the download field string into fields and traversals for constellation"""
        return [self._parse_traversal_from_field_string(traversal_string) for traversal_string in field_string]

    def _parse_traversal_from_field_string(self, field_string):
        """Parse a download traversal from the string format.

        E.g., turn FinancingTeams[Name] -> constellation consumable traversal
        """
        tokens = field_string.split("[", 1)
        if len(tokens) == 0:
            return None

        current_traversal = {"name": tokens[0].replace("]", "")}
        if len(tokens) == 1:
            return current_traversal

        current_traversal.update({"next": [self._parse_traversal_from_field_string(tokens[1])]})
        return current_traversal

    def _parse_results_from_response(self, response_objects, objects, traversals, field_parser):
        """Parse out the response for the requested fields and traversals"""
        results = []
        for obj in objects:
            results.append(self._parse_results_from_traversals(response_objects, obj, traversals, field_parser))
        return results

    def _parse_results_from_traversals(self, objects_info, obj, traversals, field_parser):
        """Parse out the response for the requested traversals"""
        object_info = objects_info.get(obj.id, {})
        results = []
        for traversal in traversals:
            if "name" not in traversal:
                continue
            field_name = traversal["name"]
            field_value = object_info.get(field_name)
            if field_name == "All":
                all_dict = {}
                for field_name, field_value in object_info.items():
                    all_dict[field_name] = field_parser.parse_field_value(
                        object_info.get("type"), field_name, field_value
                    )
                return [all_dict]

            if "next" in traversal:
                sub_traversals = traversal["next"]
                if isinstance(field_value, list):
                    results.append(
                        [
                            self._parse_results_from_traversals(
                                objects_info,
                                field_parser.to_link(field_obj),
                                sub_traversals,
                                field_parser,
                            )
                            for field_obj in field_value
                        ][0]
                    )
                else:
                    results.append(
                        self._parse_results_from_traversals(
                            objects_info,
                            field_parser.to_link(field_value),
                            sub_traversals,
                            field_parser,
                        )[0]
                    )
            else:
                results.append(field_parser.parse_field_value(object_info.get("type"), field_name, field_value))

        return results

    # Downloading cloud files
    def download_cloud_file(self, cloud_file):
        """Download the supplied cloud file to client.download_directory()"""
        file_name, extension, file_info = self.download(cloud_file, ["FileName", "FileType", "CloudFile"])
        extension = extension.replace('"', "")
        if not file_name:
            file_name = "EmeraldCloudFile_{}".format(cloud_file.id.replace(":", "_"))
        target_path = self._get_temp_file_path(file_name, extension)
        target_path = self._download_cloud_file(
            file_info["CloudFileId"], file_info["Bucket"], file_info["Key"], target_path
        )

        cloud_file.local_path = target_path
        return target_path

    def _get_temp_file_path(self, filename, extension):
        """Get a temp file path as close to filename as possible"""
        index = 0
        target_path = os.path.join(self.download_directory(), f"{filename}.{extension}")
        while os.path.exists(target_path):
            target_path = os.path.join(self.download_directory(), f"{filename}_{index}.{extension}")
            index += 1

        return target_path

    def blobsign_download(self, cloud_file_id, cloud_file_bucket, cloud_file_key):
        """Return the url to download the supplied cloud file"""
        return self._send_and_validate_request(
            "/blobsign/download",
            {
                "CloudFileId": cloud_file_id,
                "Bucket": cloud_file_bucket,
                "Key": cloud_file_key,
            },
            "Url",
        )

    def _download_cloud_file(self, cloud_file_id, cloud_file_bucket, cloud_file_key, target_path):
        """Download a cloud file based off the supplied identifiers"""
        url = self.blobsign_download(cloud_file_id, cloud_file_bucket, cloud_file_key)
        if not url:
            raise ConstellationMissingFieldException("When downloading cloud file", "url")

        # Now download the file from s3
        target_file = open(target_path, "wb")
        download_response = requests.get(url)
        target_file.write(download_response.content)
        target_file.close()

        return target_path

    def download_blob_ref(self, blob_ref, target_path=""):
        """Download a blob ref to the supplied target path"""
        target_path = target_path or self._get_temp_file_path(blob_ref.file_hash, "tmp")

        cloud_file_bucket = blob_ref.bucket
        cloud_file_key = blob_ref.key()
        if not cloud_file_bucket:
            raise ConstellationMissingFieldException("When downloading cloud file", "bucket")
        if not cloud_file_key:
            raise ConstellationMissingFieldException("When downloading cloud file", "hash")

        target_path = self._download_cloud_file("", cloud_file_bucket, cloud_file_key, target_path)
        blob_ref.local_path = target_path
        return target_path

    def upload(self, object_type, object_id, new_field_values):
        """Performs a very simple upload of a list of fields on a single object.

        Pass None for object_id to create a new object."""
        request_body = {
            "object": {"id": object_id, "type": object_type},
            "fields": new_field_values,
        }
        if object_id is None:
            # If you have no object id because you're creating a new object,
            # you need to format it slightly differently
            request_body = {"type": object_type, "fields": new_field_values}
        responses = self._send_and_validate_request(
            "/obj/upload", {"requests": [request_body]}, "responses", method="PUT"
        )

        if len(responses) != 1:
            raise ConstellationInvalidResponseException("Expected exactly 1 response from /obj/upload", responses)

        return responses[0]

    # Uploading cloud files
    def upload_cloud_file(self, file_path):
        """Uploads the given file to constellation"""
        md5 = md5_hash_file(file_path)
        extension = pathlib.Path(file_path).suffix
        if extension:
            extension = extension[1:]  # Remove the dot
        with open(file_path, "rb") as file:
            cloud_file_args = CloudFileArgs(
                file=file,
                name=pathlib.Path(file_path).stem,
                extension=extension,
                path=file_path,
                size=os.path.getsize(file_path),
                key=f"{md5.hexdigest()}.{extension}",
                content_md5=base64.b64encode(md5.digest()).decode(),
                content_disposition=f'attachment; filename="{file_path}"',
            )

            if cloud_file_args.size < self._MULTIPART_CLOUDFILE_THRESHOLD:  # 100MB
                return self._single_sign_and_upload_file_to_s3(cloud_file_args)
            return self._multi_part_sign_and_upload_file_to_s3(cloud_file_args)

    @_retry_requests_decorator
    def _single_sign_and_upload_file_to_s3(self, cloud_file_args):
        blobsign_response = self._blobsign("upload", cloud_file_args)
        headers = {
            "Content-Md5": cloud_file_args.content_md5,
            "Content-Disposition": cloud_file_args.content_disposition,
        }
        cloud_file_args.file.seek(0)
        s3_response = requests.put(
            url=blobsign_response["Url"],
            data=cloud_file_args.file,
            headers=headers,
        )
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from constellation endpoint {s3_response}: {s3_response.text}",
                s3_response.status_code,
            )
        return self._upload_cloud_file_to_constellation(blobsign_response["Key"], cloud_file_args)

    @_retry_requests_decorator
    def _upload_cloud_file_to_constellation(self, s3_key, cloud_file_args):
        cloud_field_values = {
            "$Type": "__JsonEmeraldCloudFile__",
            "Bucket": "emeraldsci-ecl-blobstore-stage",
            "CloudFileId": "None",
            "Key": s3_key,
        }
        fields = {
            "FileName": cloud_file_args.name,
            "FileSize": f'Quantity[{cloud_file_args.size}, "bytes"]',
            "FileType": f'"{cloud_file_args.extension}"',
            "CloudFile": cloud_field_values,
        }
        return self.upload("Object.EmeraldCloudFile", None, fields)

    def _blobsign(self, sign_type, cloud_file_args: CloudFileArgs, part=0, upload_id=""):
        """Return the url to upload the supplied cloud file"""
        sign_response = self._send_and_validate_request(
            "/blobsign/" + sign_type,
            {
                "Key": cloud_file_args.key,
                "ContentMD5": cloud_file_args.content_md5,
                "ContentDisposition": cloud_file_args.content_disposition,
                "Part": part,
                "UploadID": upload_id,
            },
            additional_headers={"ContentDisposition": cloud_file_args.content_disposition},
        )
        return sign_response

    def _multi_part_sign_and_upload_file_to_s3(self, cloud_file_args):
        self.s3_client = boto3.client("s3")
        upload_id, sharded_key = self._send_create_multi_part_upload_request(cloud_file_args)
        try:
            parts = self._upload_all_parts(upload_id, cloud_file_args)
            self._send_complete_multi_part_upload_request(upload_id, parts, cloud_file_args)
        except Exception as cloudFileUploadException:
            self._send_abort_multi_part_upload_request(upload_id, cloud_file_args)
            raise ConstellationException(
                "Unexpected exception while uploading cloud file", HTTPStatus.INTERNAL_SERVER_ERROR
            ) from cloudFileUploadException
        return self._upload_cloud_file_to_constellation(sharded_key, cloud_file_args)

    @_retry_requests_decorator
    def _send_create_multi_part_upload_request(self, cloud_file_args):
        blobsign_response = self._blobsign("init_multi_upload", cloud_file_args)
        s3_response = requests.post(
            url=blobsign_response["Url"],
        )
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from s3 endpoint {s3_response}: {s3_response.text}, "
                "failed to init multi-part upload",
                s3_response.status_code,
            )
        parsed_response = xmltodict.parse(s3_response.content)
        return (
            parsed_response["InitiateMultipartUploadResult"]["UploadId"],
            parsed_response["InitiateMultipartUploadResult"]["Key"],
        )

    @_retry_requests_decorator
    def _send_complete_multi_part_upload_request(self, upload_id, parts, cloud_file_args):
        blobsign_response = self._blobsign("complete_multi_upload", cloud_file_args, upload_id=upload_id)
        s3_response = requests.post(url=blobsign_response["Url"], data=parts_to_xml(parts))
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from s3 endpoint {s3_response}: {s3_response.text}, "
                "failed to complete multi-part upload",
                s3_response.status_code,
            )

    @_retry_requests_decorator
    def _send_abort_multi_part_upload_request(self, upload_id, cloud_file_args):
        blobsign_response = self._blobsign("abort_multi_upload", cloud_file_args)
        s3_response = requests.delete(url=blobsign_response["Url"])
        if s3_response.status_code != HTTPStatus.OK:
            rollbar.report_message(
                "Failed to abort multi part upload. "
                f"UploadID: {upload_id}, Key: {blobsign_response['Key']}, "
                f"S3 response: {s3_response} {s3_response.text}",
                "error",
            )
            raise ConstellationException(
                f"Unexpected non-200 response from s3 {s3_response}: {s3_response.text}, "
                "failed to abort multi part upload",
                s3_response.text,
            )

    def _upload_all_parts(self, upload_id, cloud_file_args):
        current_part_number = 1
        parts = []
        chunk = cloud_file_args.file.read(self._UPLOAD_PART_SIZE)
        while chunk != b"":
            part = self._upload_part(upload_id, current_part_number, chunk, cloud_file_args)
            parts.append(part)
            current_part_number += 1
            chunk = cloud_file_args.file.read(self._UPLOAD_PART_SIZE)
        return parts

    @_retry_requests_decorator
    def _upload_part(self, upload_id, part_number, data_chunk, cloud_file_args):
        blobsign_response = self._blobsign("upload_part", cloud_file_args, part=part_number, upload_id=upload_id)
        s3_response = requests.put(
            url=blobsign_response["Url"],
            data=data_chunk,
        )
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from s3 endpoint {s3_response}: {s3_response.text}, "
                "failed to upload part",
                s3_response.status_code,
            )
        return {"ETag": s3_response.headers["ETag"], "PartNumber": part_number}

    # Getting history of changes to an object
    def object_log(
        self,
        object_ids=None,
        types=None,
        start_date=None,
        end_date=None,
        max_results=1000,
    ):
        """Find changes on the supplied types after the supplied start_date"""
        payload = {
            "limit": max_results,
            "order": "dsc",
        }
        if object_ids is not None:
            payload["object"] = [{"id": x} for x in object_ids]
        if types is not None:
            payload["Types"] = types
        if start_date is not None:
            payload["startDate"] = start_date
        if end_date is not None:
            payload["endDate"] = end_date
        return self._send_and_validate_request(
            "/obj/objectlog",
            payload,
        )

    # Getting type information
    def get_type(self, type_name):
        """Retrieve all of the fields and their formats for a given type"""
        return self._send_and_validate_request(
            "/obj/type/{}".format(type_name),
            {},
            method="GET",
        )

    # Internal helper functions
    def _auth_header_from_token(self, auth_token):
        """Build a authorization header when an auth token is supplied"""
        return {"Authorization": "Bearer " + str(auth_token)}

    def _send_and_validate_request(
        self,
        path,
        data,
        field_to_retrieve=None,
        method="POST",
        requires_auth_token=True,
        additional_headers=None,
    ):
        """Send a request and validate a request to the given path"""
        if requires_auth_token and not self._auth_token:
            raise ConstellationNotLoggedInException()

        url = self._host + path
        if additional_headers is None:
            additional_headers = {}
        headers = {"Content-type": "application/json", **additional_headers}
        if self._auth_token:
            headers.update(self._auth_header_from_token(self._auth_token))
        if method == "POST":
            response = requests.post(
                url=url,
                data=json.dumps(data),
                headers=headers,
            )
        elif method == "PUT":
            response = requests.put(
                url=url,
                data=json.dumps(data),
                headers=headers,
            )
        elif method == "GET":
            response = requests.get(url=url, headers=headers)
        else:
            raise ConstellationInvalidMethodException(method, path)

        if response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from constellation endpoint {url}: {response.text}",
                response.status_code,
            )

        if field_to_retrieve:
            field_value = response.json().get(field_to_retrieve)
            if not field_value:
                raise ConstellationInvalidResponseException(
                    f"Could not retrieve field {field_to_retrieve} from request to {path}",
                    response.text,
                )
            return field_value

        return response.json()



import datetime
import hashlib


def constellation_date_string(dt):
    """Return a datetime formatted in the format constellation
    wants as a string like: 2021-02-17T15:16:36Z

    Note - the datetime object must be in utc time."""
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def constellation_date_from_string(date_string):
    """Return a datetime object from the string format that constellation
    uses like: 2021-02-09T04:12:02.618767Z

    Note - the datetime object will be in utc time."""
    # Handle both second and subsecond precision on datetimes
    try:
        return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        pass

    return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")


def parts_to_xml(parts):
    part_nodes = "\n".join(
        f"""
        <Part>
            <ETag>{part["ETag"]}</ETag>
            <PartNumber>{part["PartNumber"]}</PartNumber>
        </Part>
        """
        for part in parts
    )

    xml = f"""
        <?xml version="1.0" encoding="UTF-8"?>
        <CompleteMultipartUpload xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
        {part_nodes}
        "</CompleteMultipartUpload>"
    """

    return xml.encode("ascii")


def md5_hash_file(file_path):
    with open(file_path, "rb") as file:
        chunk_size = 1_000_000  # 1MB, a bit arbitrary
        md5 = hashlib.md5()
        chunk = file.read(chunk_size)
        while chunk:
            md5.update(chunk)
            chunk = file.read(chunk_size)
        return md5



import datetime
import json
import logging
import os
import pathlib
import random
import string
from zoneinfo import ZoneInfo

import boto3
import pytest

from pyecl import Constellation
from pyecl.exceptions import ConstellationObjectDoesNotExistException
from pyecl.models import Object, VariableUnitValue


def test_login(client):
    """Test that you can successfully login"""
    assert client.is_logged_in()


def test_me(client):
    """Test that the me endpoint works"""
    assert client.me() == {
        "Email": "service+manifold@emeraldcloudlab.com",
        "EmailAddress": "service+manifold@emeraldcloudlab.com",
        "Id": "id:n0k9mG8070Mk",
        "Type": "Object.User.Emerald.Developer",
        "Username": "service+manifold",
    }


# Download features


def test_simple_download(client):
    """Test a simple download on a single object"""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "Name") == "bsmith"


def test_multiple_objects_download(client):
    """Test a simple download on multiple objects"""
    assert client.download([Object("id:Z1lqpMzvkGMV"), Object("id:o1k9jAGmo6Jm")], "Name") == [
        "bsmith",
        "pavan.shah",
    ]


def test_multiple_field_download(client):
    """Test a simple download on multiple fields"""
    assert client.download(Object("id:Z1lqpMzvkGMV"), ["Name", "FirstName"]) == [
        "bsmith",
        "Ben",
    ]


def test_multiple_field_multiple_object_download(client):
    """Test a simple download multiple objects and fields"""
    assert client.download([Object("id:Z1lqpMzvkGMV"), Object("id:o1k9jAGmo6Jm")], ["Name", "FirstName"]) == [
        ["bsmith", "Ben"],
        ["pavan.shah", "Pavan"],
    ]


def test_download_non_existant_object(client):
    """Test a simple download on a non-existant object"""
    with pytest.raises(ConstellationObjectDoesNotExistException):
        client.download(Object("id:abc123"), "Name")


def test_download_non_existant_field(client):
    """Test a simple download on a non-existant field"""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FakeField") is None


def test_download_multiple_link_field(client):
    """Test a simple download on a non-existant field"""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams") == [Object("id:1ZA60vLeXa7a")]


def test_download_all(client):
    """Test downloading all fields on an object"""
    all_dict = client.download(Object("id:Z1lqpMzvkGMV"))
    assert all_dict["Name"] == "bsmith"


# Downloading links


def test_chained_download(client):
    """Test that you can chain downloads off of links"""
    assert client.download(client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams"), "Name") == ["Engineering"]


def test_single_traversal(client):
    """Test that you can traverse links"""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams[Name]") == ["Engineering"]


def test_multiple_traversal(client):
    """Test that you can perform multiple link traversals in a single request"""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams[BillingAddress[Name]]") == ["ECL-1"]


# Downloading various types


def test_download_integer(client):
    """Test that integer field types are correctly formatted"""
    assert client.download(Object("id:BYDOjvG4l3Ol"), "InjectionIndex") == 28


def test_download_string(client):
    """Test that string field types are correctly formatted"""
    assert client.download(Object("id:BYDOjvG4l3Ol"), "DataType") == "Analyte"


def test_download_quantity(client):
    """Test that quantity field types are correctly formatted"""
    assert client.download(Object("id:BYDOjvG4l3Ol"), "SampleVolume") == VariableUnitValue(10, "Microliters")


def test_download_date(client):
    """Test that date fields are correctly formatted"""
    # Note - retrieving the default units automatically is not yet supported
    assert client.download(Object("id:BYDOjvG4l3Ol"), "DateInjected") == datetime.datetime(2022, 1, 7, 17, 12, 19)


def test_file_download_with_automatic_download(client):
    """Test that files are automatically downloaded when requested"""
    cloud_file = client.download(Object("id:Z1lqpMzvkGMV"), "PhotoFile", auto_download_cloud_files=True)
    assert os.path.exists(cloud_file.local_path)


def test_file_download_without_automatic_download(client):
    """Test that files are not automatically downloaded when not requested"""
    cloud_file = client.download(Object("id:Z1lqpMzvkGMV"), "PhotoFile")
    assert cloud_file == Object("id:KBL5Dvw60ve7")
    assert cloud_file.local_path == ""


def test_blob_ref_download(client):
    """Test that blob refs are downloaded automatically correctly"""
    absorbance_data = client.download(Object("id:BYDOjvG4l3Ol"), "Absorbance")
    assert len(absorbance_data) == 7201
    assert len(absorbance_data[3]) == 2
    assert absorbance_data[3][0].unit == "'Minutes'"
    assert absorbance_data[3][1].unit == "'Milli' 'AbsorbanceUnit'"


def test_quantity_array_download(client):
    """Test that quantity arrays are downloaded correctly"""
    scatter_info = client.download(Object("id:BYDOjvG4l3Ol"), "Scattering")
    assert isinstance(scatter_info, list)
    assert len(scatter_info) == 361


def test_variable_unit_download(client):
    """Test that variable units are downloaded correctly"""
    composition = client.download(Object("id:O81aEB16GlJ1"), "Composition")
    assert isinstance(composition, list)
    assert len(composition) == 2
    assert isinstance(composition[0], list)
    assert len(composition[0]) == 2
    assert isinstance(composition[0][0], VariableUnitValue)
    assert composition[0][1] == Object("id:E8zoYvN6m61A")


def test_association_download(client):
    """Test that associations are downloaded correctly"""
    composition = client.download(Object("id:XnlV5jKZwmp3"), "ResolvedOptions")
    assert isinstance(composition, dict)
    assert len(composition) == 295
    assert isinstance(composition["AbsorbanceSamplingRate"], VariableUnitValue)
    assert composition["AbsorbanceSamplingRate"].value == 20
    assert composition["Instrument"] == Object("id:wqW9BP4ARZVw")


def test_single_upload_cloud_file(client, sample_file):
    """Create a dummy file and upload it as a cloud file"""

    file_name, contents = sample_file
    assert contents

    logging.info(f"uploading file {file_name} to cloud...")
    logging.info(contents)

    response = client.upload_cloud_file(file_name)
    resolved_object = response["resolved_object"]

    download_response = client.download(
        Object(resolved_object["id"]), ["id", "FileSize", "FileName", "FileType", "type"]
    )
    assert download_response[0] == resolved_object["id"]
    assert download_response[1] == VariableUnitValue(len(contents), "Bytes")
    assert download_response[2] == pathlib.Path(file_name).stem
    assert download_response[3] == '"txt"'
    assert download_response[4] == "Object.EmeraldCloudFile"


@pytest.fixture(scope="session")
def client():
    """Helper function to retrieve an actively logged in client

    This uses AWS secrets manager to retrieve credentials.  If you do not have access to
    ECL'S AWS secrets manager, you must override this function in order to run the tests.
    """
    secrets_client = boto3.session.Session().client(service_name="secretsmanager")
    credentials_secret = secrets_client.get_secret_value(SecretId="manifold-service-user-constellation-credentials")
    credentials_dict = json.loads(credentials_secret["SecretString"])
    username = credentials_dict["username"]
    password = credentials_dict["password"]

    # Now, get an auth token by logging into constellation for each environment
    client = Constellation(host="https://constellation-stage.emeraldcloudlab.com")
    client.login(username, password)
    return client


@pytest.fixture(scope="session")
def sample_file():
    test_file_path = f"cloud-file-test_{''.join(random.choices(string.ascii_lowercase, k=12))}.txt"

    entropy = "".join(random.choices(string.printable, k=30))
    contents = "\n".join(
        [
            "cloud file test!",
            f"generated on: {datetime.datetime.utcnow().replace(tzinfo=ZoneInfo('UTC')).isoformat()}",
            f"here's some entropy: {entropy}",
        ]
    )
    with open(test_file_path, "w") as f:
        f.write(contents)

    yield test_file_path, contents

    try:
        os.remove(test_file_path)
    except FileNotFoundError:
        pass



import base64
import json
import struct
import zlib
from dataclasses import dataclass
from typing import BinaryIO

from .exceptions import (
    BinaryObjectRepresetationUnknownTokenType,
    VariableUnitValueUnparsableExpressionException,
)
from .utils import constellation_date_from_string


class Object:
    def __init__(self, object_id, object_type="", local_path=""):
        self.id = object_id
        self.type = object_type

        # The local path is set only on EmeraldCloudFile objects that have
        # been downloaded locally
        self.local_path = local_path

    def __str__(self):
        if self.type:
            return 'Object[{}, "{}"]'.format(self.sll_style_type(), self.id)
        return 'Object["{}"]'.format(self.id)

    def __repr__(self):
        return self.__str__()

    def sll_style_type(self):
        type_string = self.type.replace("Object.", "").replace(".", "[")
        return type_string + "]" * type_string.count("]")

    def __eq__(self, o):
        if self.id != o.id:
            return False
        if self.type != "" and o.type != "":
            return self.type == o.type
        return True


class BlobRef:
    def __init__(self, bucket, path, file_hash, local_path=""):
        self.bucket = bucket
        self.path = path
        self.file_hash = file_hash

        # The local path is set only on BlobRefs that have been downloaded locally
        self.local_path = local_path

    def key(self):
        return self.path + self.file_hash

    def __str__(self):
        return "BlobRef[{}, {}, {}]".format(self.bucket, self.path, self.file_hash)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, o):
        return self.bucket == o.bucket and self.key() == o.key()

    @classmethod
    def fromDict(cls, data):
        return BlobRef(data.get("bucket"), data.get("path"), data.get("hash"))


class ConstellationFieldDefinition:
    """A structure for holding field definitions from constellation field summaries"""

    def __init__(self, count, format, subfields):
        self.count = count
        self.format = format
        self.subfields = subfields

    @classmethod
    def fromDict(cls, data):
        """Create a ConstellationFieldDefinition from a dictionary"""
        return ConstellationFieldDefinition(
            data.get("Count"),
            data.get("Format"),
            [ConstellationSubFieldDefinition.fromDict(subfieldData) for subfieldData in data.get("SubFieldSummaries")],
        )

    def field_info(self, index=-1, name=""):
        """Return the info (e.g., type, units, etc.) for the supplied index and name"""
        for subfield in self.subfields:
            if index == -1 or subfield.index == index:
                if name == "" or subfield.name == name:
                    return subfield
        return None


class ConstellationSubFieldDefinition:
    """A structure for holding sub field definitions from constellation field summaries"""

    def __init__(self, ecl_type, unit="", name="", index=0):
        self.name = name
        self.index = index
        self.unit = unit
        self.ecl_type = ecl_type

    @classmethod
    def fromDict(cls, data):
        """Create a ConstellationSubFieldDefinition from a dictionary"""
        return ConstellationSubFieldDefinition(
            data.get("Type"),
            data.get("Unit"),
            data.get("Name"),
            data.get("Index"),
        )


class ConstellationFieldParser:
    """Responsible for parsing various constellation field types"""

    def __init__(self, blob_ref_download_api=None):
        self._field_summaries_by_type = {}
        self._cloud_files_encountered = []
        self._blob_ref_download_api = blob_ref_download_api

    def add_summaries_for_type(self, summaries, ecl_type):
        """Add one or more summaries for fields of the supplied types.

        If there is an existing field summary, these will overwrite
        fields that match, but field summaries that are not passed here
        will be untouched in the existing summary store."""
        if ecl_type not in self._field_summaries_by_type:
            self._field_summaries_by_type[ecl_type] = {}
        for field_name, field_summary in summaries.items():
            self._field_summaries_by_type[ecl_type][field_name] = ConstellationFieldDefinition.fromDict(field_summary)

    def parse_field_value(self, ecl_type, field_name, field_value):
        """Parse the supplied field value into something that is python consumable"""

        # Start with trying to figure out the type from the summaries
        field_info = self._field_info_from_summary(ecl_type, field_name)
        if field_info:
            return self._parse_field_value(field_value, field_info)

        # If you couldn't read the summary, check the field value its self
        field_info = self._infer_field_type_from_value(field_value)
        if field_info:
            return self._parse_field_value(field_value, field_info)

        # If you couldn't infer it, then just return it unchanged
        return field_value

    def cloud_files_encountered(self):
        """Return a list of the cloud files that were encountered during the parsing"""
        return self._cloud_files_encountered

    def _field_info_from_summary(self, ecl_type, field_name):
        """Determine the field type (if possible) from the field summaries"""
        return self._field_summaries_by_type.get(ecl_type, {}).get(field_name)

    def _infer_field_type_from_value(self, field_value):
        """Try to infer the field type from the value"""
        subfield_values = []
        if isinstance(field_value, list):
            subfield_values = [subfield_value for subfield_value in field_value]
        else:
            subfield_values = [field_value]

        format = "Single"
        if len(subfield_values) > 1:
            # There's no way to infer the difference between a multiple field
            # and a named/indexed single field, so we assume its multiple
            format = "Multiple"

        subfield_definitions = []
        for subfield_value in subfield_values:
            if isinstance(subfield_value, dict):
                supplied_type = field_value.get("$Type")
                if supplied_type:
                    if supplied_type == "__JsonLink__":
                        subfield_definitions.append(ConstellationSubFieldDefinition("Link"))
                        continue
                    if supplied_type == "__JsonBlobReference__":
                        subfield_definitions.append(ConstellationSubFieldDefinition("BlobRef"))
                        continue
            subfield_definitions.append(ConstellationSubFieldDefinition("Unknown"))

        return ConstellationFieldDefinition(len(subfield_values), format, subfield_definitions)

    def _parse_field_value(self, field_value, field_info, map_children=True):
        """Parse the field value correctly based on the info"""
        # If we couldn't figure out the field info, then just return the value unchanged
        if not field_info or len(field_info.subfields) == 0:
            return field_value

        # If its a multiple field, then map over it (making sure to not infinitely recurse)
        if field_info.format == "Multiple" and map_children:
            return [self._parse_field_value(subvalue, field_info, map_children=False) for subvalue in field_value]

        if len(field_info.subfields) == 1:
            # If its a normal single field, then just return it parsed
            return self._parse_subfield(field_value, field_info.subfields[0])

        # If we're here, the subfield is named or indexed, so map the subfield values
        # to the subfield definitions
        if isinstance(field_value, list) and len(field_value) == len(field_info.subfields):
            results = []
            for subfield_value, subfield_info in zip(field_value, field_info.subfields):
                results.append(self._parse_subfield(subfield_value, subfield_info))
            return results

        # If we're here, then we weren't able to get the subfield definitions, so
        # return the value unchanged
        return field_value

    def _parse_subfield(self, field_value, field_info):
        """Parse a subfield value into something consumable by python"""
        # See if we know how to process this type
        if field_info.ecl_type == "Date":
            return constellation_date_from_string(field_value)
        if field_info.ecl_type == "Float":
            if field_info.unit:
                return VariableUnitValue(field_value, field_info.unit.replace('"', ""))
            return field_value
        if field_info.ecl_type == "String":
            return field_value
        if field_info.ecl_type == "Expression":
            return field_value
        if field_info.ecl_type == "Link":
            obj = self.to_link(field_value)
            if obj.type == "Object.EmeraldCloudFile":
                self._cloud_files_encountered.append(obj)
            return obj
        if field_info.ecl_type == "BlobRef":
            blob_ref = BlobRef.fromDict(field_value)
            if self._blob_ref_download_api:
                self._blob_ref_download_api(blob_ref)
                if blob_ref.local_path:
                    parsed_result = self.parse_local_file(blob_ref.local_path)
                    if parsed_result:
                        return parsed_result
            return blob_ref
        if field_info.ecl_type == "Byte":
            # If something is a byte field, its very likely to be a quantity array,
            # but not guaranteed. We'll need to investigate how to determine this 100%
            return self.parse_variable_unit_field(field_value)
        if field_info.ecl_type == "VariableUnit":
            return VariableUnitValue.parse_mathematica_expression(field_value.get("mathematica_value", ""))

        # If we don't, then just return the value unchanged
        return field_value

    def to_link(self, field_value):
        """Convert a dictionary field value to a link"""
        return Object(
            field_value.get("object", {}).get("id"),
            field_value.get("object", {}).get("type"),
        )

    def decompress_field(self, data):
        """Decompress the value of a compressed field"""

        # The data in a compressed field is first compressed with zlib, then base64 encoded,
        # and the proceeded with a 1. See:
        # https://mathematica.stackexchange.com/questions/104660/what-algorithm-do-the-compress-and-uncompress-functions-use for more info  # noqa: E501
        base64_data = data[2:]
        compressed_data = base64.b64decode(base64_data)
        encoded_data = zlib.decompress(compressed_data)

        # This is now in a special format called the binary object representation.
        # Let's parse it out
        return BinaryObjectRepresentation(encoded_data).toHumanReadableString()

    def parse_variable_unit_field(self, data):
        """Decompress and parse the value of a variable unit field"""
        decompressed_data = self.decompress_field(data)
        return VariableUnitValue.parse_mathematica_expression(decompressed_data)

    def parse_local_file(self, local_path):
        """Attempt to parse a local file into python types.

        If this fails, it will return None"""
        if not local_path:
            # If local file is not downloaded, it cannot be read
            return None

        # Blob refs are confusing because they are part ascii (specifically
        # a single line header) followed by binary data.  So we do something
        # weird where we read the file as binary and convert the first line
        # into ascii so we can read the header info
        with open(local_path, "rb") as file:
            # Note - this is an inefficient way to read the file that requires two
            # passes.  To make this more efficient, you could peek at the start of the file
            # until you hit the terminating "}" of the header, and then start reading the
            # byte contents after that.  But this is already quite fast, so I haven't
            # done it yet.
            filecontents = file.read()
            end_of_header_index = 0
            for character in filecontents:
                # 0 happens when we hit the end of the ascii data
                if character == 0:
                    break
                end_of_header_index += 1

            # the header is utf-8, but the data is not
            header = filecontents[:end_of_header_index].decode("utf-8")

            # convert the header into json
            header_data = json.loads(header)
            file.seek(end_of_header_index)

            # Validate the header is something we can read
            if header_data.get("version") != 1 or header_data.get("encoder") != "binary":
                return None

            # Pull out the unit information, which is just base 64 encoded in the
            # header:
            units = base64.b64decode(header_data.get("mathematicaUnits"))
            parsed_units = VariableUnitValue.parse_mathematica_units_from_expression(units)

            # Parse the values in the file
            results = []
            for row_values in self._parse_binary_value(file, header_data.get("binaryRowFormat")):
                if len(row_values) == 0:
                    continue
                if len(row_values) != len(parsed_units):
                    return None
                row_results = []
                for value, unit in zip(row_values, parsed_units):
                    row_results.append(VariableUnitValue(value, unit))
                results.append(row_results)

        return results

    def _parse_binary_value(self, filestream, binary_row_format):
        """Parse binary values from a filestream according to the supplied binary row format"""

        # Figure out the data width and format based on the binary row format
        # The possible types are stored in the helpfile for ?BinaryRead in mathematica
        # under more options.  Basically, we need to define a map of how many bytes
        # each type is and what the corresponding python value is for each type
        data_width = 0  # size of the row in bytes
        data_format = "<"  # MM is always little endean
        for entry_format in binary_row_format:
            (
                entry_data_width,
                entry_format,
            ) = self._width_and_format_for_mathematica_binary_type(entry_format)
            if entry_format == "" or entry_data_width == 0:
                return None
            data_width += entry_data_width
            data_format += entry_format

        while True:
            data = filestream.read(data_width)
            if len(data) < data_width:
                # end of file
                return
            yield struct.unpack(data_format, data)

    def _width_and_format_for_mathematica_binary_type(self, entry_format):
        """Calculate the width (in bytes) and the correct python type for each mathematica type.

        See ?BinaryRead (under details and options) for more info on MM types.
        See https://docs.python.org/3/library/struct.html#format-characters
        for more details on python types
        """
        width_and_format_table = {
            "Byte": [1, "b"],
            "Character8": [1, "c"],
            "Character16": [2, ""],
            "Character32": [4, ""],
            "Complex64": [8, ""],
            "Complex128": [16, ""],
            "Complex256": [32, ""],
            "Integer8": [1, "b"],
            "Integer16": [2, "h"],
            "Integer24": [3, ""],
            "Integer32": [4, "i"],
            "Integer64": [8, "q"],
            "Integer128": [16, ""],
            "Real32": [4, "f"],
            "Real64": [8, "d"],
            "Real128": [16, ""],
            "TerminatedString": [0, ""],
            "UnsignedInteger8": [1, "B"],
            "UnsignedInteger16": [2, "H"],
            "UnsignedInteger24": [3, ""],
            "UnsignedInteger32": [4, "I"],
            "UnsignedInteger64": [8, "Q"],
            "UnsignedInteger128": [16, ""],
        }
        width, struct_format = width_and_format_table.get(entry_format, [0, ""])
        return width, struct_format


class VariableUnitValue:
    """Represents a value with a convertable unit (E.g., 1.2 mm)"""

    def __init__(self, value, unit):
        self.value = value
        self.unit = unit

    def __repr__(self):
        return str(self.value) + " " + self.unit

    def __eq__(self, o):
        return isinstance(o, VariableUnitValue) and self.value == o.value and self.unit == o.unit

    @classmethod
    def parse_mathematica_expression(cls, mm_expression):
        """Convert the provided MM expression into python structures"""
        return cls._recursively_parse_mm_expression(mm_expression)

    @classmethod
    def parse_mathematica_units_from_expression(cls, mm_expression):
        """Convert a mathematica expression into one or more units"""
        parsed_expression = cls._recursively_parse_mm_expression(mm_expression)
        return cls._parse_mathematica_units_from_parsed_expression(parsed_expression)

    @classmethod
    def _parse_mathematica_units_from_parsed_expression(cls, parsed_expression):
        """Recursively parse mathematica units from a parsed expression"""
        # Once you've hit a string, you have a real unit
        if isinstance(parsed_expression, str):
            return parsed_expression

        # Map over lists, being sure to accomadate "times" units
        if isinstance(parsed_expression, list):
            # Skip empty lists
            if len(parsed_expression) == 0:
                return None
            # Ignore the IndepdentUnit operation:
            if parsed_expression[0] == "IndependentUnit":
                return parsed_expression[1]

            # Perform the times operation
            if parsed_expression[0] == "Times":
                multiplied_unit = ""
                for subunit in parsed_expression[1:]:
                    parsed_subunit = cls._parse_mathematica_units_from_parsed_expression(subunit)
                    if parsed_subunit:
                        if multiplied_unit:
                            multiplied_unit += " "
                        multiplied_unit += parsed_subunit
                return multiplied_unit

            # For normal lists, just map the result,
            # filtering None results and dropping the list keyword
            if parsed_expression[0] == "List":
                result = []
                for subunit in parsed_expression[1:]:
                    parsed_subunit = cls._parse_mathematica_units_from_parsed_expression(subunit)
                    if parsed_subunit:
                        result.append(parsed_subunit)
                return result

        return None

    @classmethod
    def _recursively_parse_mm_expression(cls, mm_expression):
        """Recursively parse a supplied mm expression into python structures"""
        if isinstance(mm_expression, bytes):
            mm_expression = mm_expression.decode("utf-8")

        mm_expression = mm_expression.strip()

        # Check if its the opening of a list
        if mm_expression.startswith("QuantityArray[") and mm_expression.endswith("]"):
            return cls._recursively_parse_mm_expression(mm_expression[14:-1])

        # Check if its the beginning of structured data
        if mm_expression.startswith("StructuredArray`StructuredData[") and mm_expression.endswith("]"):
            return cls._parse_structured_data_expression(mm_expression[31:-1])

        # Check if its the beginning of simple structured array
        if mm_expression.startswith("StructuredArray[") and mm_expression.endswith("]"):
            return cls._parse_structured_array_expression(mm_expression[16:-1])

        # Check if its a quanity
        if mm_expression.startswith("Quantity[") and mm_expression.endswith("]"):
            return cls._parse_quantity_expression(mm_expression[9:-1])

        # Check if its a rule list opening via List[Rule[ keyword
        if mm_expression.startswith("List[Rule[") and mm_expression.endswith("]"):
            return cls._parse_mm_rule_list_expression(mm_expression[5:-1])

        # Check if its a list opening via List keyword
        if mm_expression.startswith("List[") and mm_expression.endswith("]"):
            return cls._parse_mm_list_expression(mm_expression[5:-1])

        # Check if its a list opening via [ character
        if mm_expression.startswith("[") and mm_expression.endswith("]"):
            return cls._parse_mm_list_expression(mm_expression[1:-1])

        # Check if its a list opening via {}
        if mm_expression.startswith("{") and mm_expression.endswith("}"):
            return cls._parse_mm_list_expression(mm_expression[1:-1])

        # Check if its a Rule
        if mm_expression.startswith("Rule[") and mm_expression.endswith("]"):
            return cls._parse_mm_rule_expression(mm_expression[5:-1])

        # Check if its an object
        if mm_expression.startswith("Object[") and mm_expression.endswith("]"):
            return cls._parse_mm_object_expression(mm_expression[7:-1])

        # Check if its a MM null:
        if mm_expression == "Null":
            return None

        # Check if its an int
        try:
            return int(mm_expression.replace("`", ""))
        except ValueError:
            pass

        # Check if its a float
        try:
            return float(mm_expression.replace("`", ""))
        except ValueError:
            pass

        # At this point, its a symbol, so just return it as a string
        # we'll drop the enclosing quotes as they're not needed
        if mm_expression.startswith('"') and mm_expression.endswith('"'):
            mm_expression = mm_expression[1:-1]
        return mm_expression

    @classmethod
    def _parse_mm_list_expression(cls, mm_expression):
        """Parse a list mm expression into a python structure"""
        components = cls._split_by_unenclosed_substring(mm_expression, ",", {"{": "}", "[": "]"})
        return [cls._recursively_parse_mm_expression(component) for component in components]

    @classmethod
    def _parse_mm_rule_expression(cls, mm_expression):
        """Parse a rule mm expression into a python dict"""
        components = cls._split_by_unenclosed_substring(mm_expression, ",", {"{": "}", "[": "]"})
        if len(components) == 2:
            return {
                cls._recursively_parse_mm_expression(components[0]): cls._recursively_parse_mm_expression(components[1])
            }
        return {}

    @classmethod
    def _parse_mm_rule_list_expression(cls, mm_expression):
        """Parse a rule list mm expression into a python dict"""
        components = cls._split_by_unenclosed_substring(mm_expression, ",", {"{": "}", "[": "]"})
        results = {}
        for component in components:
            result = cls._recursively_parse_mm_expression(component)
            if isinstance(result, dict):
                results.update(result)
        return results

    @classmethod
    def _parse_quantity_expression(cls, mm_expression):
        """Parse a quantity mm expression into a python structure"""
        components = cls._parse_mm_list_expression(mm_expression)
        if len(components) != 2:
            raise VariableUnitValueUnparsableExpressionException("Unexpected quantity components", components)

        return VariableUnitValue(components[0], components[1])

    @classmethod
    def _parse_mm_object_expression(cls, mm_expression):
        """Parse a list mm expression into a python structure"""
        components = cls._split_by_unenclosed_substring(mm_expression.replace(" ", ""), ",", {"{": "}", "[": "]"})
        return Object(object_id=components[-1], object_type=".".join(components[:-1]))

    @classmethod
    def _parse_structured_data_expression(cls, mm_expression):
        """Parse a structured data mm expression into a python structure"""
        # The structure data looks like:
        # {Length}, {{val1, ..., valn}, {unit1, ..., unitN}, {additional info}}
        # where each thing can also be a multiple if there's multiple values
        # we don't care about thet length or additional info,
        # so we're just going to pull the values and unit
        structured_data_components = cls._parse_mm_list_expression(mm_expression)

        # there's two forms (at least) that MM can put this in, so let's check each
        if len(structured_data_components) == 2:
            value_list = structured_data_components[1]
            if len(value_list) < 2:
                raise VariableUnitValueUnparsableExpressionException("Unexpected value list", value_list)

            all_values = value_list[0]
            all_units = value_list[1]
        elif len(structured_data_components) > 2:
            all_values = structured_data_components[1]
            all_units = structured_data_components[2]
        else:
            raise VariableUnitValueUnparsableExpressionException(
                "Unexpected structured data components", structured_data_components
            )

        # Sometimes the values are at the top level, and sometimes they're one level
        # down - thanks MM!
        is_multiple = True
        if len(all_values) == 0 or not isinstance(all_values[0], list):
            all_values = [all_values]
            is_multiple = False

        variable_unit_values = []
        for values in all_values:
            # Convert the units and values into variable unit objects
            variable_unit_subvalues = []
            for value, unit in zip(values, all_units):
                variable_unit_subvalues.append(VariableUnitValue(value, unit))
            if is_multiple:
                variable_unit_values.append(variable_unit_subvalues)
            else:
                variable_unit_values = variable_unit_subvalues
        return variable_unit_values

    @classmethod
    def _parse_structured_array_expression(cls, mm_expression):
        """Parse a structured array mm expression into a python structure"""
        components = cls._parse_mm_list_expression(mm_expression)
        if len(components) != 3:
            raise VariableUnitValueUnparsableExpressionException("Unexpected structured array components", components)

        return components[2]

    @classmethod
    def _find_unenclosed_substring(cls, string, substring, enclosing_chars):
        """Find, for example, a comma not contained in {}"""
        open_enclosing_chars = []
        for index in range(len(string)):
            if len(open_enclosing_chars) == 0 and string[index:].startswith(substring):
                return index
            for opening_char, closing_char in enclosing_chars.items():
                if string[index] == opening_char:
                    open_enclosing_chars.append(opening_char)
                if string[index] == closing_char:
                    open_enclosing_chars.remove(opening_char)
        return -1

    @classmethod
    def _split_by_unenclosed_substring(cls, string, substring, enclosing_chars):
        """Split, for example, a comma not contained in {}"""
        results = []
        index = cls._find_unenclosed_substring(string, substring, enclosing_chars)
        while index != -1:
            results.append(string[:index])
            string = string[index + 1 :]
            index = cls._find_unenclosed_substring(string, substring, enclosing_chars)
        results.append(string)
        return results


class BinaryObjectRepresentation:
    """The binary object representation used internally by mathematica.

    This is primarily useful for dealing with compressed fields"""

    def __init__(self, encoded_data):
        self._encoded_data = encoded_data

    def toHumanReadableString(self):
        """Return a human readable string version of the encoded data"""
        return self._parse_binary_object_representation()

    def _parse_binary_object_representation(self):
        """Parse the MM binary object representation into a human readable string.

        See https://mathematica.stackexchange.com/questions/104660/what-algorithm-do-the-compress-and-uncompress-functions-use for more info  # noqa: E501
        """
        expression = ""
        data = self._encoded_data
        while len(data) != 0:
            new_expression, data = self._parse_token_from_binary_object_represenation(data)
            expression = expression + str(new_expression)

        return expression

    def _parse_token_from_binary_object_represenation(self, data):
        """Parse a single token from the binary object representation"""

        # First, check if we're seeing the header
        if len(data) >= 4 and data[:4] == b"!boR":
            # Just strip the header
            return "", data[4:]

        # If not, then check if we can read the leading character to determine the token type
        if len(data) > 0:
            token_type_char = data[0]
            data = data[1:]
            if token_type_char == 105:  # b"i"
                return self._parse_integer_token(data[:4]), data[4:]
            if token_type_char == 114:  # b"r"
                return self._parse_real_token(data[:8]), data[8:]
            if token_type_char == 83:  # b"S"
                return self._retrieve_and_parse_string_token(data)
            if token_type_char == 115:  # b"s"
                # Note that strings and symbols parse exactly the same way
                return self._retrieve_and_parse_string_token(data)
            if token_type_char == 73:  # b"I"
                # Note large integers are parsed the same as strings
                return self._retrieve_and_parse_string_token(data)
            if token_type_char == 102:  # b"f"
                return self._retrieve_and_parse_expression_token(data)
            if token_type_char == 101:  # b"e"
                return self._retrieve_and_parse_matrix_token(data)

            raise BinaryObjectRepresetationUnknownTokenType(token_type_char, data)

        # If data was empty, then just return nothing
        return "", b""

    def _parse_integer_token(self, bytes):
        """Parse up to 4 little endian bytes as an integer"""
        return struct.unpack("<i", bytes)[0]

    def _parse_real_token(self, bytes):
        """Parse real number from the supplied bytes"""
        return struct.unpack("<d", bytes)[0]

    def _retrieve_and_parse_string_token(self, bytes):
        """Pull the string token out and parse it, returning the string and
        the remainder of the bytes"""
        # the first four bytes hold the size of the string
        string_size = self._parse_integer_token(bytes[:4])
        string_bytes = bytes[4:]
        return string_bytes[:string_size].decode("utf-8"), string_bytes[string_size:]

    def _retrieve_and_parse_expression_token(self, bytes):
        """Pull the expression token out and parse it, returning the expression and
        the remainder of the bytes"""
        # the first four bytes hold the number of tokens you should pull in order to
        # form the expression
        expression_size = self._parse_integer_token(bytes[:4])
        expression_bytes = bytes[4:]

        # First, parse out the symbol head
        (
            symbol_head,
            expression_bytes,
        ) = self._parse_token_from_binary_object_represenation(expression_bytes)

        # Now, parse the arguments of the head
        arguments = []
        while expression_size > 0 and len(expression_bytes) > 0:
            (
                argument,
                expression_bytes,
            ) = self._parse_token_from_binary_object_represenation(expression_bytes)
            arguments.append(str(argument))
            expression_size = expression_size - 1

        return symbol_head + "[" + ", ".join(arguments) + "]", expression_bytes

    def _retrieve_and_parse_matrix_token(self, bytes):
        """Pull the matrix token out and parse it, returning the expression and
        the remainder of the bytes"""
        matrix_dimensionality = self._parse_integer_token(bytes[:4])
        expression_bytes = bytes[4:]

        dimension_sizes = []
        while len(dimension_sizes) < matrix_dimensionality:
            dimension_sizes.append(self._parse_integer_token(expression_bytes[:4]))
            expression_bytes = expression_bytes[4:]

        return self._parse_matrix_real_values(dimension_sizes, expression_bytes)

    def _parse_matrix_real_values(self, dimension_sizes, expression_bytes):
        """Recursively parse a matrix of reals from the supplied bytes"""
        if len(dimension_sizes) == 0:
            return self._parse_real_token(expression_bytes[:8]), expression_bytes[8:]

        dimension_size = dimension_sizes[0]
        dimension_sizes = dimension_sizes[1:]
        results = []
        while len(results) < dimension_size:
            subresult, expression_bytes = self._parse_matrix_real_values(dimension_sizes, expression_bytes)
            results.append(str(subresult))

        return "{" + ", ".join(results) + "}", expression_bytes


@dataclass
class CloudFileArgs:
    """Class for arguments that are needed to upload a cloud file to s3 and constellation"""

    file: BinaryIO
    name: str
    extension: str
    path: str
    size: int
    key: str
    content_md5: str
    content_disposition: str
